 /***********************************************************
  *  SIMPLE HD44780 BASED LCD DRIVER - Modified by Fabrizio May 2001
  *
  *  LCD is connected, in 4 bit mode, on PORT A as follows:
  *  (no busy flag check is performed)
  *
  *  PA7      PA6  PA5  PA4  PA3  PA2  PA1  PA0
  *backligth  R/W   E    RS   D7   D6   D5   D4
  ************************************************************/
#include "lcd.h"  // functions declaration


/*-------------------------------------------------
*   delay= p x 5 uS  at 8 MHz
*--------------------------------------------------*/
void lcd_delay(unsigned int p) 
{
  unsigned int i;
  byte j;
  for(i=0;i<p;i++)
    for (j=0;j<10;j++)
	   asm("nop");	//CLK=8 MHz-->j<10
}

/*-------------------------------------------------
*               _______
*  ENABLE (PA5)        |________
*--------------------------------------------------*/
void toggle_E()
{
  outp(inp(PORTA) | 0x20,PORTA);        // E=1
  outp(inp(PORTA) & 0xDF,PORTA);        // E=0
}

/*-------------------------------------------------
 *  Display initialization + sets always write mode
 *--------------------------------------------------*/
void lcd_init(void)
{
  outp(0xFF,DDRA);
  cbi(PORTA,6);		// always write (0)
  lcd_delay(10000);
  outp(0x03,PORTA); toggle_E(); lcd_delay(500);
  outp(0x03,PORTA); toggle_E(); lcd_delay(100);
  outp(0x03,PORTA); toggle_E(); lcd_delay(100);
  outp(0x02,PORTA); toggle_E(); lcd_delay(100);
  outp(0x02,PORTA); toggle_E();
  outp(0x08,PORTA); toggle_E(); lcd_delay(100);
  outp(0x00,PORTA); toggle_E();
  outp(0x0F,PORTA); toggle_E(); lcd_delay(100);
  outp(0x00,PORTA); toggle_E();
  outp(0x06,PORTA); toggle_E(); lcd_delay(100);
  lcd_clr();
}

/*-------------------------------------------------
*   sends a byte of data to display in 4 bit mode
*--------------------------------------------------*/
void send(byte data,int rs)
{
	byte tmp;
	if (rs==1) outp(inp(PORTA)|0x10,PORTA);	//out data
	else outp(inp(PORTA)& 0xEF,PORTA);	//out instruction

	tmp=(inp(PORTA)& 0xF0)|((data & 0xF0)>>4); //MSB of data
	outp(tmp,PORTA); toggle_E();
	tmp=(inp(PORTA)& 0xF0)|(data & 0x0F); //LSB of data
	outp(tmp,PORTA); toggle_E();
}

/*-------------------------------------------------
 *  Clear display
 *--------------------------------------------------*/
void lcd_clr()
{
  send(0x01,0);	//sends 00000001 to LCD as instruction
  lcd_delay(1000);
}

/*-------------------------------------------------
 *  Home position
 *--------------------------------------------------*/
void lcd_home()
{
  send(0x02,0);	//sends 00000010 to LCD as instruction
  lcd_delay(1000);
}

/*-------------------------------------------------
 *  LCD & Cursor control (Blinking, ON/OFF,...)
 *--------------------------------------------------*/
void lcd_control(byte disonoff, byte curonoff, byte curblink)
{ byte temp;
  temp=0x08;
  if (disonoff==1) temp|=0x04;
  if (curonoff==1) temp|=0x02;
  if (curblink==1) temp|=0x01;
  send(temp,0);
  lcd_delay(100);
}

/*-------------------------------------------------
 *  Goto position   (0x00 - 1.row, 1.col)
 *                  (0x40 - 1.row, 2.col)
 *--------------------------------------------------*/
void lcd_goto(byte pos)
{
  send(0x80|pos,0);  
  lcd_delay(100);
}

/*-------------------------------------------------
 *  Put character on LCD
 *--------------------------------------------------*/
void lcd_putch(byte data)
{ 
  send(data,1);	// send a data to LCD
  lcd_delay(100);
}

/*-------------------------------------------------
 *  Display null terminated string
 *--------------------------------------------------*/
void lcd_putstr(byte *data)
{
    while(*data) {
        lcd_putch(*data);
        data++;
    }
}

/*-------------------------------------------------
 *   Display 0-99
 * --------------------------------------------------*/
void print2(unsigned int x)
{
  unsigned int y;
  if (x<100)
    {
      y=x/10;lcd_putch(y+0x30);x-=(y*10);
      lcd_putch(x+0x30);
    }
  else lcd_putstr("Err");
}

/*-------------------------------------------------
 *  Display 0-999 unsigned char or int (8 or 16bit)
 *--------------------------------------------------*/
void print3(unsigned int x)
{
  unsigned int y;
  if (x<1000)
  {
    y=x/100;lcd_putch(y+0x30);x-=(y*100);
    y=x/10;lcd_putch(y+0x30);x-=(y*10);
    lcd_putch(x+0x30);
  }
  else lcd_putstr("Err");
}

/*-------------------------------------------------
 *   Simple printf function (no fp, and strings)
 *--------------------------------------------------*/
int dprintf(const char *format, ...){
  static const char hex[] = "0123456789ABCDEF";
  char format_flag;
  unsigned int u_val, div_val, base;
  char *ptr;
  va_list ap;

  va_start (ap, format);
  for (;;) {
     while ((format_flag = *format++) != '%') {      // Until '%' or '\0'
       if (!format_flag) {
         va_end (ap);
         return (0);
       }
       lcd_putch (format_flag);
     }

     switch (format_flag = *format++) {
       case 'c': format_flag = va_arg(ap,int);
       default:  lcd_putch(format_flag); continue;

       case 'd': base = 10; div_val = 10000; goto CONVERSION_LOOP;
       case 'x': base = 16; div_val = 0x10;

       CONVERSION_LOOP:
       u_val = va_arg(ap,int);
       if (format_flag == 'd') {
         if (((int)u_val) < 0) {
           u_val = - u_val;
           lcd_putch ('-');
         }
         while (div_val > 1 && div_val > u_val) div_val /= 10;
       }
       do {
         lcd_putch (hex[u_val / div_val]);
         u_val %= div_val;
         div_val /= base;
       } while (div_val);
    }
  }
}

